# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the CLI functionality for update command.

Author(s): Alec Delaney
"""

import os
import pathlib
import shutil
import threading

from click.testing import CliRunner

import circfirm
import circfirm.backend.cache
import tests.helpers
from circfirm.cli import cli

RUNNER = CliRunner()

ORIGINAL_VERSION = "6.0.0"


@tests.helpers.as_circuitpy
def test_update() -> None:
    """Test the update command when in CIRCUITPY mode."""
    try:
        tests.helpers.set_firmware_version(ORIGINAL_VERSION)

        threading.Thread(target=tests.helpers.wait_and_set_bootloader).start()
        result = RUNNER.invoke(cli, ["update", "--language", "cs"])
        expected_version = "6.1.0"
        expected_uf2_filename = circfirm.backend.cache.get_uf2_filename(
            "feather_m4_express", expected_version, language="cs"
        )
        expected_uf2_filepath = tests.helpers.get_mount_node(expected_uf2_filename)
        assert result.exit_code == 0
        assert os.path.exists(expected_uf2_filepath)
        os.remove(expected_uf2_filepath)

    finally:
        board_folder = circfirm.backend.cache.get_board_folder("feather_m4_express")
        if board_folder.exists():
            shutil.rmtree(board_folder)


@tests.helpers.as_circuitpy
def test_update_pre_release() -> None:
    """Tests the update command when in CIRCUITPY mode with the pre-release flag."""
    try:
        tests.helpers.set_firmware_version(ORIGINAL_VERSION)

        threading.Thread(target=tests.helpers.wait_and_set_bootloader).start()
        result = RUNNER.invoke(cli, ["update", "--language", "cs", "--pre-release"])
        expected_version = "6.2.0-beta.2"
        expected_uf2_filename = circfirm.backend.cache.get_uf2_filename(
            "feather_m4_express", expected_version, language="cs"
        )
        expected_uf2_filepath = tests.helpers.get_mount_node(expected_uf2_filename)
        assert result.exit_code == 0
        assert os.path.exists(expected_uf2_filepath)
        os.remove(expected_uf2_filepath)

    finally:
        board_folder = circfirm.backend.cache.get_board_folder("feather_m4_express")
        if board_folder.exists():
            shutil.rmtree(board_folder)


@tests.helpers.as_bootloader
def test_update_bootloader_mode() -> None:
    """Tests the update command when in bootloader mode."""
    try:
        expected_version = "6.1.0"
        board_id = "feather_m4_express"
        result = RUNNER.invoke(
            cli, ["update", "--board-id", board_id, "--language", "cs"]
        )
        expected_uf2_filename = circfirm.backend.cache.get_uf2_filename(
            board_id, expected_version, language="cs"
        )
        expected_uf2_filepath = tests.helpers.get_mount_node(expected_uf2_filename)
        assert result.exit_code == 0
        assert os.path.exists(expected_uf2_filepath)
        os.remove(expected_uf2_filepath)

    finally:
        board_folder = circfirm.backend.cache.get_board_folder("feather_m4_express")
        if board_folder.exists():
            shutil.rmtree(board_folder)


@tests.helpers.as_circuitpy
def test_update_to_lower() -> None:
    """Tests the update command when the current version is higher."""
    tests.helpers.set_firmware_version("100.0.0")

    threading.Thread(target=tests.helpers.wait_and_set_bootloader).start()
    result = RUNNER.invoke(cli, ["update", "--language", "cs"])

    mount_path = pathlib.Path(tests.helpers.get_mount())
    mount_uf2_files = list(mount_path.rglob("*.uf2"))

    assert result.exit_code == 0
    assert not mount_uf2_files



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Cross-platform script for deleting folders in the CI/Makefile.

Author(s): Alec Delaney
"""

# pragma: no cover

import shutil
import sys

target = sys.argv[1]
shutil.rmtree(target)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the CLI's config command functionality.

Author(s): Alec Delaney
"""

import json

import yaml
from click.testing import CliRunner

from circfirm.cli import cli

RUNNER = CliRunner()


def get_printed_default_settings() -> None:
    """Get the default (template) settings as printed."""
    with open("circfirm/templates/settings.yaml", encoding="utf-8") as yamlfile:
        settings = yaml.safe_load(yamlfile)
    return f"{json.dumps(settings, indent=4)}\n"


def test_config() -> None:
    """Tests the config view command."""
    expected_output = get_printed_default_settings()

    # Test viewing all settings
    result = RUNNER.invoke(cli, ["config", "view"])
    assert result.exit_code == 0
    assert result.output == expected_output

    # Test viewing specific setting
    result = RUNNER.invoke(cli, ["config", "view", "output.supporting.silence"])
    assert result.exit_code == 0
    assert result.output == "false\n"

    # Test viewing non-existent setting
    result = RUNNER.invoke(cli, ["config", "view", "doesnotexist"])
    assert result.exit_code != 0

    # Test writing a setting
    RUNNER.invoke(cli, ["config", "edit", "output.supporting.silence", "true"])
    result = RUNNER.invoke(cli, ["config", "view", "output.supporting.silence"])
    assert result.exit_code == 0
    assert result.output == "true\n"
    RUNNER.invoke(cli, ["config", "edit", "output.supporting.silence", "false"])
    result = RUNNER.invoke(cli, ["config", "view", "output.supporting.silence"])
    assert result.exit_code == 0
    assert result.output == "false\n"

    # Test writing a non-existent setting
    result = RUNNER.invoke(cli, ["config", "edit", "doesnotexist", "123"])
    assert result.exit_code != 0

    # Test writing to setting tree
    result = RUNNER.invoke(cli, ["config", "edit", "output", "123"])
    assert result.exit_code != 0

    # Test writing bad type
    RUNNER.invoke(cli, ["config", "edit", "output.supporting.silence", "123"])
    assert result.exit_code != 0


def test_config_reset() -> None:
    """Tests the reseting of the config settings file."""
    RUNNER.invoke(cli, ["config", "edit", "output.supporting.silence", "true"])
    result = RUNNER.invoke(cli, ["config", "view", "output.supporting.silence"])
    assert result.exit_code == 0
    assert result.output == "true\n"

    expected_settings = get_printed_default_settings()

    result = RUNNER.invoke(cli, ["config", "reset"])
    result = RUNNER.invoke(cli, ["config", "view"])
    assert result.exit_code == 0
    assert result.output == expected_settings



# SPDX-FileCopyrightText: 2024 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""Configuration file for sphinx documentation."""

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sphinx_rtd_theme

project = "circfirm"
copyright = "2024, Alec Delaney"
author = "Alec Delaney"
version = release = "0.0.0+auto.0"


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_tabs.tabs",
    "sphinxcontrib.jquery",
]

templates_path = ["_templates"]
exclude_patterns = []

intersphinx_mapping = {
    "CircuitPython": ("https://docs.circuitpython.org/en/latest/", None),
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path(), "."]



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""CLI functionality for the query subcommand.

Author(s): Alec Delaney
"""

import re

import click
import requests

import circfirm
import circfirm.backend.github
import circfirm.backend.s3
import circfirm.cli


@click.group()
def cli():
    """Query things like latest versions and board lists."""


@cli.command(name="board-ids")
@click.option("-r", "--regex", default=".*", help="Regex pattern to use for board IDs")
def query_board_ids(regex: str) -> None:
    """Query the local CircuitPython board list."""
    settings = circfirm.cli.get_settings()
    gh_token = settings["token"]["github"]
    do_output = not settings["output"]["supporting"]["silence"]
    circfirm.cli.maybe_support(
        "Boards list will now be synchronized with the git repository."
    )
    if not gh_token:
        circfirm.cli.maybe_support(
            "Please note that this operation can only be performed 60 times per hour due to GitHub rate limiting."
        )
    try:
        if do_output:
            boards = circfirm.cli.announce_and_await(
                "Fetching boards list",
                circfirm.backend.github.get_board_id_list,
                args=(gh_token,),
            )
        else:
            boards = circfirm.backend.github.get_board_id_list(gh_token)
    except ValueError as err:
        raise click.ClickException(err.args[0])
    except requests.ConnectionError as err:
        raise click.ClickException(
            "Issue with requesting information from git repository, check network connection"
        )
    for board in boards:
        board_id = board.strip()
        result = re.search(regex, board_id)
        if result:
            click.echo(board_id)


@cli.command(name="versions")
@click.argument("board-id")
@click.option("-l", "--language", default="en_US", help="CircuitPython language/locale")
@click.option("-r", "--regex", default=".*", help="Regex pattern to use for versions")
def query_versions(board_id: str, language: str, regex: str) -> None:
    """Query the CircuitPython versions available for a board."""
    versions = circfirm.backend.s3.get_board_versions(board_id, language, regex=regex)
    for version in reversed(versions):
        click.echo(version)


@cli.command(name="latest")
@click.argument("board-id", default="raspberry_pi_pico")
@click.option("-l", "--language", default="en_US", help="CircuitPython language/locale")
@click.option(
    "-p",
    "--pre-release",
    is_flag=True,
    default=False,
    help="Consider pre-release versions",
)
def query_latest(board_id: str, language: str, pre_release: bool) -> None:
    """Query the latest CircuitPython versions available."""
    version = circfirm.backend.s3.get_latest_board_version(
        board_id, language, pre_release
    )
    if version:
        click.echo(version)



# SPDX-FileCopyrightText: 2022 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Pytest configuration file.

Author(s): Alec Delaney
"""

import os
import pathlib
import shutil
from typing import Union

import click
import pytest

BACKUP_FOLDER = pathlib.Path("tests/backup/")
APP_DIR = pathlib.Path(click.get_app_dir("circfirm")).resolve()

CONFIG_EXISTS = APP_DIR.exists()


def pytest_sessionstart(session: pytest.Session) -> None:
    """Save the current cron table before testing."""
    # Load environment variables if not in GitHub Actions
    if "GH_TOKEN" not in os.environ:  # pragma: no cover
        with open(".env", encoding="utf-8") as envfile:
            env_contents = envfile.read()
            for envline in env_contents.split("\n"):
                if not envline:
                    continue
                name, value = envline.split("=")
                os.environ[name] = value

    # Create the backup directory
    BACKUP_FOLDER.mkdir(exist_ok=True)

    # Save existing settings, if they exist
    if CONFIG_EXISTS:  # pragma: no cover
        shutil.move(APP_DIR, "tests/backup")


def pytest_sessionfinish(
    session: pytest.Session, exitstatus: Union[int, pytest.ExitCode]
) -> None:
    """Restore the previous cron table after testing."""
    try:
        shutil.rmtree(APP_DIR)
    except FileNotFoundError:  # pragma: no cover
        pass
    if CONFIG_EXISTS:  # pragma: no cover
        shutil.move("tests/backup/circfirm", APP_DIR.parent)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""CLI functionality for the cache subcommand.

Author(s): Alec Delaney
"""

import os
import pathlib
import shutil
import sys
from typing import Optional

import click

import circfirm
import circfirm.backend.cache
import circfirm.cli
import circfirm.startup


@click.group()
def cli():
    """Work with cached firmwares."""


@cli.command()
@click.option("-b", "--board-id", default=None, help="CircuitPython board ID")
@click.option("-v", "--version", default=None, help="CircuitPython version")
@click.option("-l", "--language", default=None, help="CircuitPython language/locale")
def clear(
    board_id: Optional[str], version: Optional[str], language: Optional[str]
) -> None:
    """Clear the cache, either entirely or for a specific board/version."""
    if board_id is None and version is None and language is None:
        shutil.rmtree(circfirm.UF2_ARCHIVE)
        circfirm.startup.ensure_app_setup()
        click.echo("Cache cleared!")
        return

    glob_pattern = "*-*" if board_id is None else f"*-{board_id}"
    language_pattern = "-*" if language is None else f"-{language}"
    glob_pattern += language_pattern
    version_pattern = "-*" if version is None else f"-{version}.uf2"
    glob_pattern += version_pattern
    matching_files = pathlib.Path(circfirm.UF2_ARCHIVE).rglob(glob_pattern)
    for matching_file in matching_files:
        matching_file.unlink()

    # Delete board folder if empty
    for board_folder in pathlib.Path(circfirm.UF2_ARCHIVE).glob("*"):
        if len(os.listdir(board_folder)) == 0:
            shutil.rmtree(board_folder)

    click.echo("Cache cleared of specified entries!")


@cli.command(name="list")
@click.option("-b", "--board-id", default=None, help="CircuitPython board ID")
def cache_list(board_id: Optional[str]) -> None:
    """List all the boards/versions cached."""
    board_list = os.listdir(circfirm.UF2_ARCHIVE)

    if not board_list:
        circfirm.cli.maybe_support("Versions have not been cached yet for any boards.")
        return

    if board_id is not None and board_id not in board_list:
        circfirm.cli.maybe_support(
            f"No versions for board '{board_id}' are not cached."
        )
        return

    specified_board = board_id if board_id is not None else None
    boards = circfirm.backend.cache.get_sorted_boards(specified_board)

    for rec_boardid, rec_boardvers in boards.items():
        click.echo(f"{rec_boardid}")
        for rec_boardver, rec_boardlangs in rec_boardvers.items():
            for rec_boardlang in rec_boardlangs:
                click.echo(f"  * {rec_boardver} ({rec_boardlang})")


@cli.command(name="save")
@click.argument("board-id")
@click.argument("version")
@click.option("-l", "--language", default="en_US", help="CircuitPython language/locale")
def cache_save(board_id: str, version: str, language: str) -> None:
    """Install a version of CircuitPython to cache."""
    try:
        circfirm.cli.announce_and_await(
            f"Caching firmware version {version} for {board_id}",
            circfirm.backend.cache.download_uf2,
            args=(board_id, version, language),
        )
    except ConnectionError as err:
        raise click.exceptions.ClickException(err.args[0])



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
# Based on code from circlink (Authored by Alec Delaney, licensed under MIT license)
#
# SPDX-License-Identifier: MIT

"""CLI functionality for the config subcommand.

Author(s): Alec Delaney
"""

import json
import os

import click
import yaml

import circfirm
import circfirm.cli
import circfirm.startup


@click.group()
def cli():
    """View and update the configuration settings for the circfirm CLI."""


@cli.command(name="view")
@click.argument("setting", default="all")
def config_view(setting: str) -> None:
    """View a config setting."""
    # Get the settings, show all settings if no specific on is specified
    settings = circfirm.cli.get_settings()
    if setting == "all":
        click.echo(json.dumps(settings, indent=4))
        return

    # Get the specified settings
    config_args = setting.split(".")
    try:
        for extra_arg in config_args[:-1]:
            settings = settings[extra_arg]
        value = settings[config_args[-1]]
    except KeyError:
        raise click.ClickException(f"Setting {setting} does not exist")

    # Show the specified setting
    click.echo(json.dumps(value, indent=4))


@cli.command(name="edit")
@click.argument("setting")
@click.argument("value")
def config_edit(
    setting: str,
    value: str,
) -> None:
    """Update a config setting."""
    # Get the settings, use another reference to parse
    orig_settings = circfirm.cli.get_settings()
    target_setting = orig_settings
    config_args = setting.split(".")

    # Handle bool conversions
    if value.lower() == "true":
        value = True
    elif value.lower() == "false":
        value = False

    # Attempt to parse for the specified config setting and set it
    try:
        for extra_arg in config_args[:-1]:
            target_setting = target_setting[extra_arg]
        prev_value = target_setting[config_args[-1]]
        prev_value_type = type(prev_value)
        if prev_value_type == dict:
            raise ValueError
        if prev_value_type == bool and value not in (True, False):
            raise TypeError
        target_setting[config_args[-1]] = prev_value_type(value)
    except KeyError:
        raise click.ClickException(f"Setting {setting} does not exist")
    except TypeError:
        raise click.ClickException(
            f"Cannot use that value for this setting, must be of type {prev_value_type}"
        )
    except ValueError:
        raise click.ClickException(
            "Cannot change this setting, please change the sub-settings within it"
        )

    # Write the settings back to the file
    with open(circfirm.SETTINGS_FILE, mode="w", encoding="utf-8") as yamlfile:
        yaml.safe_dump(orig_settings, yamlfile)


@cli.command(name="reset")
def config_reset() -> None:
    """Reset the configuration file with the provided template."""
    os.remove(circfirm.SETTINGS_FILE)
    src, dest = [
        (src, dest)
        for src, dest in circfirm.startup.TEMPLATE_LIST
        if dest == circfirm.SETTINGS_FILE
    ][0]
    circfirm.startup.ensure_template(src, dest)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Backend functionality for the working with CircuitPython GitHub repository.

Author(s): Alec Delaney
"""

import datetime
import re
from typing import List, Tuple, TypedDict

import requests

BASE_REQUESTS_HEADERS = {
    "Accept": "application/vnd.github+json",
    "X-GitHub-Api-Version": "2022-11-28",
}

BOARDS_REGEX = r"ports/.+/boards/([^/]+)"


class RateLimit(TypedDict):
    """Format of a rate limit dictionary."""

    limit: int
    remaining: int
    reset: int
    used: int
    resource: str


class GitTreeItem(TypedDict):
    """Format of a git tree item dictionary."""

    path: str
    mode: str
    type: str
    size: int
    sha: str
    url: str


def get_rate_limit() -> Tuple[int, int, datetime.datetime]:
    """Get the rate limit for the GitHub REST endpoint."""
    response = requests.get(
        url="https://api.github.com/rate_limit",
        headers=BASE_REQUESTS_HEADERS,
    )
    limit_info: RateLimit = response.json()["rate"]
    available: int = limit_info["remaining"]
    total: int = limit_info["limit"]
    reset_time = datetime.datetime.fromtimestamp(limit_info["reset"])
    return available, total, reset_time


def get_board_id_list(token: str) -> List[str]:
    """Get a list of CircuitPython boards."""
    boards = set()
    headers = BASE_REQUESTS_HEADERS.copy()
    if token:
        headers["Authorization"] = f"Bearer {token}"
    response = requests.get(
        url="https://api.github.com/repos/adafruit/circuitpython/git/trees/main",
        params={
            "recursive": True,
        },
        headers=headers,
    )
    try:
        tree_items: List[GitTreeItem] = response.json()["tree"]
    except KeyError as err:
        raise ValueError("Could not parse JSON response, check token") from err
    for tree_item in tree_items:
        if tree_item["type"] != "tree":
            continue
        result = re.match(BOARDS_REGEX, tree_item["path"])
        if result:
            boards.add(result[1])
    return sorted(boards)



# SPDX-FileCopyrightText: 2022 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Basic startup functionality for the tools.

Author(s): Alec Delaney
"""

import os
import pathlib
import shutil
from typing import List, Tuple

import click

FOLDER_LIST: List[str] = []
FILE_LIST: List[str] = []
TEMPLATE_LIST: List[Tuple[str, str]] = []


def specify_app_dir(app_name: str) -> str:
    """Set up the application directory."""
    app_path = click.get_app_dir(app_name)
    FOLDER_LIST.append(app_path)
    return app_path


def specify_folder(*path_parts: str) -> str:
    """Add a folder to the global record list."""
    folder_path = os.path.join(*path_parts)
    FOLDER_LIST.append(folder_path)
    return folder_path


def specify_file(*path_parts: str) -> str:
    """Add a file to the global record list."""
    file_path = os.path.join(*path_parts)
    FILE_LIST.append(file_path)
    return file_path


def specify_template(src_path: str, dest_path: str) -> str:
    """Add a template to the global record list."""
    TEMPLATE_LIST.append((src_path, dest_path))
    return dest_path


def ensure_dir(dir_path: str, /) -> None:
    """Ensure the directory exists, create if needed."""
    if not os.path.exists(dir_path):
        os.makedirs(dir_path, exist_ok=True)


def _ensure_file(file_path: str, /) -> None:
    """Ensure the file exists, touch to create if needed."""
    if not os.path.exists(file_path):
        pathlib.Path(file_path).touch(exist_ok=True)


def ensure_template(src_path: str, dest_path: str, /) -> None:
    """Ensure the template exists, copy it if needed."""
    if not os.path.exists(dest_path):
        shutil.copyfile(src_path, dest_path)


def ensure_app_setup() -> None:
    """Ensure the entire application folder is set up."""
    for folder in FOLDER_LIST:
        ensure_dir(folder)
    for file in FILE_LIST:
        _ensure_file(file)
    for template in TEMPLATE_LIST:
        ensure_template(*template)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the CLI's query command functionality.

Author(s): Alec Delaney
"""

import os
from typing import NoReturn

import pytest
import requests
from click.testing import CliRunner

import circfirm.backend.github
import tests.helpers
from circfirm.cli import cli

RUNNER = CliRunner()


def simulate_no_connection(arg: str) -> NoReturn:
    """Simulate a network error by raising requests.ConnectionError."""
    raise requests.ConnectionError


def test_query_board_ids(monkeypatch: pytest.MonkeyPatch) -> None:
    """Tests the ability to query the boards using the CLI."""
    # Test an unauthenticated request with supporting text
    board_ids = tests.helpers.get_board_ids_from_git()
    pre_expected_output = "".join([f"{board}\n" for board in board_ids])
    expected_output = "\n".join(
        [
            "Boards list will now be synchronized with the git repository.",
            "Please note that this operation can only be performed 60 times per hour due to GitHub rate limiting.",
            "Fetching boards list... done",
            pre_expected_output,
        ]
    )

    result = RUNNER.invoke(cli, ["query", "board-ids"])
    assert result.exit_code == 0
    assert result.output == expected_output

    # Test an authenticated request without supporting text
    result = RUNNER.invoke(
        cli, ["config", "edit", "token.github", os.environ["GH_TOKEN"]]
    )
    assert result.exit_code == 0
    result = RUNNER.invoke(cli, ["config", "edit", "output.supporting.silence", "true"])
    assert result.exit_code == 0
    result = RUNNER.invoke(cli, ["query", "board-ids"])
    assert result.exit_code == 0
    assert result.output == pre_expected_output

    # Test a request with a faulty token
    result = RUNNER.invoke(cli, ["config", "edit", "token.github", "badtoken"])
    assert result.exit_code == 0
    result = RUNNER.invoke(cli, ["query", "board-ids"])
    assert result.exit_code != 0

    result = RUNNER.invoke(cli, ["config", "reset"])
    assert result.exit_code == 0

    # Tests failure when cannot fetch results due to no network connection
    monkeypatch.setattr(
        circfirm.backend.github, "get_board_id_list", simulate_no_connection
    )
    result = RUNNER.invoke(cli, ["query", "board-ids"])
    assert result.exit_code != 0
    assert (
        result.output.split("\n")[-2]
        == "Error: Issue with requesting information from git repository, check network connection"
    )


def test_query_versions() -> None:
    """Tests the ability to query firmware versions using the CLI."""
    board = "adafruit_feather_rp2040"
    language = "cs"
    expected_versions = [
        "6.2.0-beta.0",
        "6.2.0-beta.1",
        "6.2.0-beta.2",
    ]
    expected_output = "".join([f"{version}\n" for version in expected_versions])

    result = RUNNER.invoke(
        cli,
        [
            "query",
            "versions",
            board,
            "--language",
            language,
        ],
    )
    assert result.exit_code == 0
    assert result.output == expected_output


def test_query_latest() -> None:
    """Tests the ability to query the latest version of the firmware using the CLI."""
    board = "adafruit_feather_rp2040"
    language = "cs"
    expected_output = "6.2.0-beta.2\n"

    # Test without pre-releases included
    result = RUNNER.invoke(
        cli,
        [
            "query",
            "latest",
            board,
            "--language",
            language,
        ],
    )
    assert result.exit_code == 0
    assert result.output == ""

    # Test with pre-releases included
    result = RUNNER.invoke(
        cli,
        ["query", "latest", board, "--language", language, "--pre-release"],
    )
    assert result.exit_code == 0
    assert result.output == expected_output



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the backend GitHub functionality.

Author(s): Alec Delaney
"""

import circfirm.backend.s3


def test_get_board_versions() -> None:
    """Tests getting firmware versions for a given board."""
    board = "adafruit_feather_rp2040"
    language = "cs"
    expected_versions = [
        "6.2.0-beta.2",
        "6.2.0-beta.1",
        "6.2.0-beta.0",
    ]
    versions = circfirm.backend.s3.get_board_versions(board, language)
    assert versions == expected_versions

    # Chck that invalid versions are skipped for code coverage
    _ = circfirm.backend.s3.get_board_versions(board, regex=r".*")



# SPDX-FileCopyrightText: 2022 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Main CLI functionality for the tool.

Author(s): Alec Delaney
"""

import importlib.util
import os
import pkgutil
import shutil
import sys
import time
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, TypeVar

import click
import click_spinner
import yaml

import circfirm
import circfirm.backend.cache
import circfirm.backend.device
import circfirm.startup

_T = TypeVar("_T")


@click.group()
@click.version_option(package_name="circfirm")
def cli() -> None:
    """Manage CircuitPython firmware from the command line."""
    circfirm.startup.ensure_app_setup()


def maybe_support(msg: str) -> None:
    """Output supporting text based on the configurable settings."""
    settings = get_settings()
    do_output: bool = not settings["output"]["supporting"]["silence"]
    if do_output:
        click.echo(msg)


def get_board_id(
    circuitpy: Optional[str], bootloader: Optional[str], board: Optional[str]
) -> Tuple[str, str]:
    """Get the board ID of a device via CLI."""
    if not board:
        if not circuitpy and bootloader:
            click.echo("CircuitPython device found, but it is in bootloader mode!")
            click.echo(
                "Please put the device out of bootloader mode, or use the --board-id option."
            )
            sys.exit(3)
        board = circfirm.backend.device.get_board_info(circuitpy)[0]

        click.echo("Board ID detected, please switch the device to bootloader mode.")
        while not (bootloader := circfirm.backend.device.find_bootloader()):
            time.sleep(1)
    return bootloader, board


def get_connection_status() -> Tuple[Optional[str], Optional[str]]:
    """Get the status of a connectted CircuitPython device as a CIRCUITPY and bootloader location."""
    circuitpy = circfirm.backend.device.find_circuitpy()
    bootloader = circfirm.backend.device.find_bootloader()
    if not circuitpy and not bootloader:
        click.echo("CircuitPython device not found!")
        click.echo("Check that the device is connected and mounted.")
        sys.exit(1)
    return circuitpy, bootloader


def ensure_bootloader_mode(bootloader: Optional[str]) -> None:
    """Ensure the connected device is in bootloader mode."""
    if not bootloader:
        if circfirm.backend.device.find_circuitpy():
            click.echo("CircuitPython device found, but is not in bootloader mode!")
            click.echo("Please put the device in bootloader mode.")
            sys.exit(2)


def download_if_needed(board: str, version: str, language: str) -> None:
    """Download the firmware for a given board, version, and language via CLI."""
    if not circfirm.backend.cache.is_downloaded(board, version, language):
        try:
            announce_and_await(
                "Downloading UF2",
                circfirm.backend.cache.download_uf2,
                args=(board, version, language),
            )
        except ConnectionError as err:
            click.echo(" failed")  # Mark as failed
            click.echo(f"Error: {err.args[0]}")
            sys.exit(4)
    else:
        click.echo("Using cached firmware file")


def copy_cache_firmware(
    board: str, version: str, language: str, bootloader: str
) -> None:
    """Copy the cached firmware for a given board, version, and language to the bootloader via CLI."""
    uf2file = circfirm.backend.cache.get_uf2_filepath(board, version, language)
    uf2filename = os.path.basename(uf2file)
    uf2_path = os.path.join(bootloader, uf2filename)
    announce_and_await(
        f"Copying UF2 to {board}", shutil.copyfile, args=(uf2file, uf2_path)
    )
    click.echo("Device should reboot momentarily.")


def announce_and_await(
    msg: str,
    func: Callable[..., _T],
    args: Iterable = (),
    kwargs: Optional[Dict[str, Any]] = None,
    *,
    use_spinner: bool = True,
) -> _T:
    """Announce an action to be performed, do it, then announce its completion."""
    if kwargs is None:
        kwargs = {}
    fmt_msg = f"{msg}..."
    spinner = click_spinner.spinner()
    click.echo(fmt_msg, nl=False)
    if use_spinner:
        spinner.start()
    try:
        try:
            resp = func(*args, **kwargs)
        finally:
            if use_spinner:
                spinner.stop()
        click.echo(" done")
        return resp
    except BaseException as err:
        click.echo(" failed")
        raise err


def get_settings() -> Dict[str, Any]:
    """Get the contents of the settings file."""
    with open(circfirm.SETTINGS_FILE, encoding="utf-8") as yamlfile:
        return yaml.safe_load(yamlfile)


def load_subcmd_folder(path: str, super_import_name: str) -> None:
    """Load subcommands dynamically from a folder of modules and packages."""
    subcmd_names = [
        (modname, ispkg) for _, modname, ispkg in pkgutil.iter_modules((path,))
    ]
    subcmd_paths = [
        os.path.abspath(os.path.join(path, subcmd_name[0]))
        for subcmd_name in subcmd_names
    ]

    for (subcmd_name, ispkg), subcmd_path in zip(subcmd_names, subcmd_paths):
        import_name = ".".join([super_import_name, subcmd_name])
        import_path = subcmd_path if ispkg else subcmd_path + ".py"
        module_spec = importlib.util.spec_from_file_location(import_name, import_path)
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        source_cli: click.MultiCommand = getattr(module, "cli")
        if isinstance(source_cli, click.Group):
            subcmd = click.CommandCollection(sources=(source_cli,))
            subcmd.help = source_cli.__doc__
        else:
            subcmd = source_cli
        cli.add_command(subcmd, subcmd_name)


# Load extra commands from the rest of the circfirm.cli subpackage
cli_pkg_path = os.path.dirname(os.path.abspath(__file__))
cli_pkg_name = "circfirm.cli"
load_subcmd_folder(cli_pkg_path, cli_pkg_name)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the CLI functionality for about command.

Author(s): Alec Delaney
"""

from click.testing import CliRunner

from circfirm.cli import cli

RUNNER = CliRunner()


def test_about() -> None:
    """Tests the about command."""
    result = RUNNER.invoke(cli, ["about"])
    assert result.exit_code == 0
    assert result.output == "Written by Alec Delaney, licensed under MIT License.\n"



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the backend device functionality.

Author(s): Alec Delaney
"""

import pytest

import circfirm.backend.device
import tests.helpers


@tests.helpers.as_circuitpy
def test_find_circuitpy() -> None:
    """Tests finding a CircuitPython device when boot_out.txt is present."""
    mount_location = tests.helpers.get_mount()
    circuitpy = circfirm.backend.device.find_circuitpy()
    assert circuitpy == mount_location


@tests.helpers.as_not_present
def test_find_circuitpy_absent() -> None:
    """Tests finding a CircuitPython device when boot_out.txt is absent."""
    circuitpy = circfirm.backend.device.find_circuitpy()
    assert circuitpy is None


@tests.helpers.as_bootloader
def test_find_bootloader() -> None:
    """Tests finding a CircuitPython device in bootloader mode when info_uf2.txt is present."""
    mount_location = tests.helpers.get_mount()
    bootloader = circfirm.backend.device.find_bootloader()
    assert bootloader == mount_location


@tests.helpers.as_not_present
def test_find_bootloader_absent() -> None:
    """Tests finding a CircuitPython device in bootloader mode when info_uf2.txt is absent."""
    bootloader = circfirm.backend.device.find_bootloader()
    assert bootloader is None
    tests.helpers.copy_uf2_info()


@tests.helpers.as_circuitpy
def test_get_board_info() -> None:
    """Tests getting the board ID and firmware version from the UF2 info file."""
    # Test successful parsing
    mount_location = tests.helpers.get_mount()
    board_id = circfirm.backend.device.get_board_info(mount_location)[0]
    assert board_id == "feather_m4_express"

    # Test unsuccessful parsing of board ID
    with open(
        tests.helpers.get_mount_node(circfirm.BOOTOUT_FILE), mode="w", encoding="utf-8"
    ) as bootfile:
        bootfile.write("junktext")
    with pytest.raises(ValueError):
        circfirm.backend.device.get_board_info(mount_location)

    # Test unsuccessful parsing of firmware version
    with open(
        tests.helpers.get_mount_node(circfirm.BOOTOUT_FILE), mode="w", encoding="utf-8"
    ) as bootfile:
        bootfile.write("junktext\nBoard ID:feather_m4_express")
    with pytest.raises(ValueError):
        circfirm.backend.device.get_board_info(mount_location)



# SPDX-FileCopyrightText: 2022 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Shared functionality for the tool.

Author(s): Alec Delaney
"""

import os

from circfirm.startup import (
    specify_app_dir,
    specify_file,
    specify_folder,
    specify_template,
)

# Folders
APP_DIR = specify_app_dir("circfirm")
UF2_ARCHIVE = specify_folder(APP_DIR, "archive")

# Files
_SETTINGS_FILE_SRC = os.path.abspath(
    os.path.join(__file__, "..", "templates", "settings.yaml")
)
SETTINGS_FILE = specify_template(
    _SETTINGS_FILE_SRC, os.path.join(APP_DIR, "settings.yaml")
)
UF2_BOARD_LIST = specify_file(APP_DIR, "boards.txt")

UF2INFO_FILE = "info_uf2.txt"
BOOTOUT_FILE = "boot_out.txt"



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the CLI functionality for current command.

Author(s): Alec Delaney
"""

from click.testing import CliRunner

import tests.helpers
from circfirm.cli import cli

RUNNER = CliRunner()


@tests.helpers.as_circuitpy
def test_current() -> None:
    """Tests the current name and version commands."""
    # Test when connected in CIRCUITPY mode
    result = RUNNER.invoke(cli, ["current", "board-id"])
    assert result.exit_code == 0
    assert result.output == "feather_m4_express\n"

    result = RUNNER.invoke(cli, ["current", "version"])
    assert result.exit_code == 0
    assert result.output == "8.0.0-beta.6\n"


@tests.helpers.as_bootloader
def test_current_in_bootloader() -> None:
    """Tests the current command whenn connected in bootloader mode."""
    result = RUNNER.invoke(cli, ["current", "board-id"])
    assert result.exit_code != 0



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""CLI functionality for the update subcommand.

Author(s): Alec Delaney
"""

from typing import Optional

import click
import packaging.version

import circfirm.backend.device
import circfirm.backend.s3
import circfirm.cli.install


@click.command()
@click.option(
    "-b",
    "--board-id",
    default=None,
    help="Assume the given board ID (and connect in bootloader mode)",
)
@click.option("-l", "--language", default="en_US", help="CircuitPython langauge/locale")
@click.option(
    "-p",
    "--pre-release",
    is_flag=True,
    default=False,
    help="Whether pre-release versions should be considered",
)
def cli(board_id: Optional[str], language: str, pre_release: bool) -> None:
    """Update a connected board to the latest CircuitPython version."""
    circuitpy, bootloader = circfirm.cli.get_connection_status()
    if circuitpy:
        _, current_version = circfirm.backend.device.get_board_info(circuitpy)
    else:
        click.echo(
            "Bootloader mode detected - cannot check the currently installed version"
        )
        click.echo(
            "The latest version will be installed regardless of the currently installed version."
        )
        current_version = "0.0.0"
    bootloader, board_id = circfirm.cli.get_board_id(circuitpy, bootloader, board_id)

    new_version = circfirm.backend.s3.get_latest_board_version(
        board_id, language, pre_release
    )
    if packaging.version.Version(current_version) >= packaging.version.Version(
        new_version
    ):
        click.echo(
            f"Current version ({current_version}) is at or higher than proposed new update ({new_version})"
        )
        return

    circfirm.cli.ensure_bootloader_mode(bootloader)
    circfirm.cli.download_if_needed(board_id, new_version, language)
    circfirm.cli.copy_cache_firmware(board_id, new_version, language, bootloader)



# SPDX-FileCopyrightText: 2022 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Shared backend functionality.

Author(s): Alec Delaney
"""

import enum


class Language(enum.Enum):
    """Avaiable languages for boards."""

    BAHASA_INDONESIAN = "ID"
    CZECH = "cs"
    GERMAN = "de_DE"
    UK_ENGLISH = "en_GB"
    US_ENGLISH = "en_US"
    PIRATE_ENGLISH = "en_x_pirate"
    SPANISH = "es"
    FILIPINO = "fil"
    FRENCH = "fr"
    HINDI = "hi"
    ITALIAN = "it_IT"
    JAPANESE = "ja"
    KOREAN = "ko"
    DUTCH = "nl"
    POLISH = "pl"
    BRAZILIAN_PORTUGESE = "pt_BR"
    RUSSIAN = "ru"
    SWEDISH = "sv"
    TURKISH = "tr"
    MANDARIN_LATIN_PINYIN = "zh_Latn_pinyin"


_ALL_LANGAGES = [language.value for language in Language]
_ALL_LANGUAGES_REGEX = "|".join(_ALL_LANGAGES)
_VALID_VERSIONS_CAPTURE = r"(\d+\.\d+\.\d+(?:-(?:\balpha\b|\bbeta\b)\.\d+)*)"
FIRMWARE_REGEX_PATTERN = "-".join(
    [
        r"adafruit-circuitpython",
        r"[board]",
        r"[language]",
        r"[version]\.uf2",
    ]
)
FIRMWARE_REGEX = (
    FIRMWARE_REGEX_PATTERN.replace(r"[board]", r"(.*)")
    .replace(r"[language]", f"({_ALL_LANGUAGES_REGEX})")
    .replace(r"[version]", _VALID_VERSIONS_CAPTURE)
)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Backend functionality for the working with the CircuitPython firmware S3 bucket.

Author(s): Alec Delaney
"""

import re
from typing import List, Optional

import boto3
import botocore
import botocore.client
import packaging.version
from mypy_boto3_s3 import S3ServiceResource

import circfirm.backend

S3_CONFIG = botocore.client.Config(signature_version=botocore.UNSIGNED)
S3_RESOURCE: S3ServiceResource = boto3.resource("s3", config=S3_CONFIG)
BUCKET_NAME = "adafruit-circuit-python"
BUCKET = S3_RESOURCE.Bucket(BUCKET_NAME)


def get_board_versions(
    board_id: str, language: str = "en_US", *, regex: Optional[str] = None
) -> List[str]:
    """Get a list of CircuitPython versions for a given board."""
    prefix = f"bin/{board_id}/{language}"
    firmware_regex = circfirm.backend.FIRMWARE_REGEX_PATTERN.replace(
        r"[board]", board_id
    ).replace(r"[language]", language)
    version_regex = f"({regex})" if regex else circfirm.backend._VALID_VERSIONS_CAPTURE
    firmware_regex = firmware_regex.replace(r"[version]", version_regex)
    s3_objects = BUCKET.objects.filter(Prefix=prefix)
    versions = set()
    for s3_object in s3_objects:
        result = re.match(f"{prefix}/{firmware_regex}", s3_object.key)
        if result:
            try:
                _ = packaging.version.Version(result[1])
                versions.add(result[1])
            except packaging.version.InvalidVersion:
                pass
    return sorted(versions, key=packaging.version.Version, reverse=True)


def get_latest_board_version(
    board_id: str, language: str, pre_release: bool
) -> Optional[str]:
    """Get the latest version for a board in a given language."""
    versions = get_board_versions(board_id, language)
    if not pre_release:
        versions = [
            version
            for version in versions
            if not packaging.version.Version(version).is_prerelease
        ]
    if versions:
        return versions[0]
    return None



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Backend functionality for the working with the connected devices.

Author(s): Alec Delaney
"""

import pathlib
import re
from typing import Optional, Tuple

import psutil

import circfirm

BOARD_ID_REGEX = r"Board ID:\s*(.*)"
BOARD_VER_REGEX = (
    r"Adafruit CircuitPython (\d+\.\d+\.\d+(?:-(?:\balpha\b|\bbeta\b)\.\d+)*)"
)


def get_board_info(device_path: str) -> Tuple[str, str]:
    """Get the attached CircuitPytho board's name and version."""
    bootout_file = pathlib.Path(device_path) / circfirm.BOOTOUT_FILE
    with open(bootout_file, encoding="utf-8") as infofile:
        contents = infofile.read()
    board_match = re.search(BOARD_ID_REGEX, contents)
    if not board_match:
        raise ValueError("Could not parse the board ID from the boot out file")
    version_match = re.search(BOARD_VER_REGEX, contents)
    if not version_match:
        raise ValueError("Could not parse the firmware version from the boot out file")
    return board_match[1], version_match[1]


def _find_device(filename: str) -> Optional[str]:
    """Find a specific connected device."""
    for partition in psutil.disk_partitions():
        try:
            bootout_file = pathlib.Path(partition.mountpoint) / filename
            if bootout_file.exists():
                return partition.mountpoint
        except PermissionError:  # pragma: no cover
            pass
    return None


def find_circuitpy() -> Optional[str]:
    """Find CircuitPython device in non-bootloader mode."""
    return _find_device(circfirm.BOOTOUT_FILE)


def find_bootloader() -> Optional[str]:
    """Find CircuitPython device in bootloader mode."""
    return _find_device(circfirm.UF2INFO_FILE)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the CLI functionality for install command.

Author(s): Alec Delaney
"""

import os
import shutil
import threading

from click.testing import CliRunner

import circfirm.backend.cache
import tests.helpers
from circfirm.cli import cli

RUNNER = CliRunner()

ERR_NOT_FOUND = 1
ERR_FOUND_CIRCUITPY = 2
ERR_IN_BOOTLOADER = 3
ERR_UF2_DOWNLOAD = 4

VERSION = "8.0.0-beta.6"


@tests.helpers.as_circuitpy
def test_install_successful() -> None:
    """Tests the successful use of the install command."""
    try:
        # Test successfully installing the firmware
        threading.Thread(target=tests.helpers.wait_and_set_bootloader).start()
        result = RUNNER.invoke(cli, ["install", VERSION])
        assert result.exit_code == 0
        expected_uf2_filename = circfirm.backend.cache.get_uf2_filename(
            "feather_m4_express", VERSION
        )
        expected_uf2_filepath = tests.helpers.get_mount_node(expected_uf2_filename)
        assert os.path.exists(expected_uf2_filepath)
        os.remove(expected_uf2_filepath)

        # Test using cached version of firmware
        result = RUNNER.invoke(
            cli, ["install", VERSION, "--board-id", "feather_m4_express"]
        )
        assert result.exit_code == 0
        assert "Using cached firmware file" in result.output
        os.remove(expected_uf2_filepath)

    finally:
        board_folder = circfirm.backend.cache.get_board_folder("feather_m4_express")
        if board_folder.exists():
            shutil.rmtree(board_folder)


@tests.helpers.as_not_present
def test_install_no_mount() -> None:
    """Tests the install command when a mounted drive is not found."""
    result = RUNNER.invoke(
        cli, ["install", VERSION, "--board-id", "feather_m4_express"]
    )
    assert result.exit_code == ERR_NOT_FOUND


@tests.helpers.as_circuitpy
def test_install_as_circuitpy() -> None:
    """Tests the install command when a mounted CIRCUITPY drive is found."""
    result = RUNNER.invoke(
        cli, ["install", VERSION, "--board-id", "feather_m4_express"]
    )
    assert result.exit_code == ERR_FOUND_CIRCUITPY


@tests.helpers.as_bootloader
def test_install_bad_version() -> None:
    """Tests the install command using a bad board version."""
    result = RUNNER.invoke(
        cli, ["install", "doesnotexist", "--board-id", "feather_m4_express"]
    )
    assert result.exit_code == ERR_UF2_DOWNLOAD

    # Test using install when in bootloader mode
    result = RUNNER.invoke(cli, ["install", VERSION])
    assert result.exit_code == ERR_IN_BOOTLOADER



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""CLI functionality for the install subcommand.

Author(s): Alec Delaney
"""

from typing import Optional

import click

import circfirm.cli


@click.command()
@click.argument("version")
@click.option("-l", "--language", default="en_US", help="CircuitPython language/locale")
@click.option(
    "-b",
    "--board-id",
    default=None,
    help="Assume the given board ID (and connect in bootloader mode)",
)
def cli(version: str, language: str, board_id: Optional[str]) -> None:
    """Install the specified version of CircuitPython."""
    circuitpy, bootloader = circfirm.cli.get_connection_status()
    bootloader, board_id = circfirm.cli.get_board_id(circuitpy, bootloader, board_id)
    circfirm.cli.ensure_bootloader_mode(bootloader)
    circfirm.cli.download_if_needed(board_id, version, language)
    circfirm.cli.copy_cache_firmware(board_id, version, language, bootloader)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the CLI functionality for cache command.

Author(s): Alec Delaney
"""

import os
import pathlib
import shutil

from click.testing import CliRunner

import circfirm
import circfirm.backend.cache
import tests.helpers
from circfirm.cli import cli

RUNNER = CliRunner()


def test_cache_list() -> None:
    """Tests the cache list command."""
    # Test empty cache
    result = RUNNER.invoke(cli, ["cache", "list"])
    assert result.exit_code == 0
    assert result.output == "Versions have not been cached yet for any boards.\n"

    # Move firmware files to app directory
    tests.helpers.copy_firmwares()

    # Get full list expected response
    with open("tests/assets/responses/full_list.txt", encoding="utf-8") as respfile:
        expected_response = respfile.read()
    result = RUNNER.invoke(cli, ["cache", "list"])
    assert result.exit_code == 0
    assert result.output == expected_response

    # Test specific board that is present
    with open(
        "tests/assets/responses/specific_board.txt", encoding="utf-8"
    ) as respfile:
        expected_response = respfile.read()
    result = RUNNER.invoke(cli, ["cache", "list", "--board-id", "feather_m4_express"])
    assert result.exit_code == 0
    assert result.output == expected_response

    # Test specific board, version, and language response
    fake_board = "does_not_exist"
    with open(
        "tests/assets/responses/specific_board.txt", encoding="utf-8"
    ) as respfile:
        expected_response = respfile.read()
    result = RUNNER.invoke(cli, ["cache", "list", "--board-id", fake_board])
    assert result.exit_code == 0
    assert result.output == f"No versions for board '{fake_board}' are not cached.\n"

    # Clean Up after test
    shutil.rmtree(circfirm.UF2_ARCHIVE)
    os.mkdir(circfirm.UF2_ARCHIVE)


def test_cache_save() -> None:
    """Tests the cache save command."""
    board = "feather_m4_express"
    version = "7.3.0"
    langauge = "fr"

    # Save a specific firmware (successful)
    result = RUNNER.invoke(
        cli, ["cache", "save", board, version, "--language", langauge]
    )
    assert result.exit_code == 0
    expected_path = circfirm.backend.cache.get_uf2_filepath(
        board, version, language=langauge
    )
    assert expected_path.exists()
    shutil.rmtree(expected_path.parent.resolve())

    # Save a specific firmware (unsuccessful)
    result = RUNNER.invoke(
        cli, ["cache", "save", board, version, "--language", "nolanguage"]
    )
    assert result.exit_code == 1
    assert result.output == (
        "Caching firmware version 7.3.0 for feather_m4_express... failed\n"
        "Error: Could not download the specified UF2 file:\n"
        "https://downloads.circuitpython.org/bin/feather_m4_express/"
        "nolanguage/"
        "adafruit-circuitpython-feather_m4_express-nolanguage-7.3.0.uf2\n"
    )


def test_cache_clear() -> None:
    """Tests the cache clear command."""
    board = "feather_m4_express"
    version = "7.1.0"
    langauge = "zh_Latn_pinyin"

    # Move firmware files to app directory
    tests.helpers.copy_firmwares()

    # Remove a specific firmware from the cache
    result = RUNNER.invoke(
        cli,
        [
            "cache",
            "clear",
            "--board-id",
            board,
            "--version",
            version,
            "--language",
            langauge,
        ],
    )
    board_folder = pathlib.Path(circfirm.UF2_ARCHIVE) / board
    uf2_file = (
        board_folder
        / "adafruit-circuitpython-feather_m4_express-zh_Latn_pinyin-7.1.0.uf2"
    )
    assert result.exit_code == 0
    assert result.output == "Cache cleared of specified entries!\n"
    assert not uf2_file.exists()
    assert board_folder.exists()

    # Remove a specific board firmware from the cache
    result = RUNNER.invoke(cli, ["cache", "clear", "--board-id", board])
    assert result.exit_code == 0
    assert result.output == "Cache cleared of specified entries!\n"
    assert not board_folder.exists()

    # Remove entire cache
    result = RUNNER.invoke(cli, ["cache", "clear"])
    assert result.exit_code == 0
    assert result.output == "Cache cleared!\n"
    assert len(list(board_folder.parent.glob("*"))) == 0



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""CLI functionality for the about subcommand.

Author(s): Alec Delaney
"""

import click


@click.command()
def cli() -> None:
    """Information about circfirm."""
    click.echo("Written by Alec Delaney, licensed under MIT License.")



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the backend cache functionality.

Author(s): Alec Delaney
"""

import pathlib
import shutil

import pytest

import circfirm.backend.cache


def test_get_board_folder() -> None:
    """Tests getting UF2 information."""
    board_id = "feather_m4_express"
    board_path = circfirm.backend.cache.get_board_folder(board_id)
    expected_path = pathlib.Path(circfirm.UF2_ARCHIVE) / board_id
    assert board_path.resolve() == expected_path.resolve()


def test_get_uf2_filepath() -> None:
    """Tests getting the UF2 filepath."""
    board_id = "feather_m4_express"
    language = "en_US"
    version = "7.0.0"

    created_path = circfirm.backend.cache.get_uf2_filepath(
        "feather_m4_express", "7.0.0", "en_US"
    )
    expected_path = (
        pathlib.Path(circfirm.UF2_ARCHIVE)
        / board_id
        / f"adafruit-circuitpython-{board_id}-{language}-{version}.uf2"
    )
    assert created_path.resolve() == expected_path.resolve()


def test_download_uf2() -> None:
    """Tests the UF2 download functionality."""
    board_id = "feather_m4_express"
    language = "en_US"
    version = "junktext"

    # Test bad download candidate
    expected_path = (
        circfirm.backend.cache.get_board_folder(board_id)
        / f"adafruit-circuitpython-{board_id}-{language}-{version}.uf2"
    )
    with pytest.raises(ConnectionError):
        circfirm.backend.cache.download_uf2(board_id, version, language)
    assert not expected_path.exists()
    assert not expected_path.parent.exists()

    # Test good download candidate
    assert not circfirm.backend.cache.is_downloaded(board_id, version)
    version = "7.0.0"
    circfirm.backend.cache.download_uf2(board_id, version, language)
    expected_path = (
        circfirm.backend.cache.get_board_folder(board_id)
        / f"adafruit-circuitpython-{board_id}-{language}-{version}.uf2"
    )
    assert expected_path.exists()
    assert circfirm.backend.cache.is_downloaded(board_id, version)

    # Clean up post tests
    shutil.rmtree(expected_path.parent)


def test_parse_firmware_info() -> None:
    """Tests the ability to get firmware information."""
    board_id = "feather_m4_express"
    language = "en_US"

    # Test successful parsing
    for version in ("8.0.0", "9.0.0-beta.2"):
        try:
            board_folder = circfirm.backend.cache.get_board_folder(board_id)
            circfirm.backend.cache.download_uf2(board_id, version, language)
            downloaded_filename = [file.name for file in board_folder.glob("*")][0]

            (
                parsed_version,
                parsed_language,
            ) = circfirm.backend.cache.parse_firmware_info(downloaded_filename)
            assert parsed_version == version
            assert parsed_language == language
        finally:
            # Clean up post tests
            shutil.rmtree(board_folder)

    # Test failed parsing
    with pytest.raises(ValueError):
        circfirm.backend.cache.parse_firmware_info("cannotparse")



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests the backend GitHub functionality.

Author(s): Alec Delaney
"""

import os
import pathlib
import shutil

import pytest

import circfirm.backend.github
import tests.helpers


def test_get_board_list() -> None:
    """Tests the ability of the backend to get the board list."""
    # Test successful retrieval
    token = os.environ["GH_TOKEN"]
    board_list = circfirm.backend.github.get_board_id_list(token)
    expected_board_list = tests.helpers.get_board_ids_from_git()
    assert board_list == expected_board_list

    # Test unsuccessful retrieval
    with pytest.raises(ValueError):
        circfirm.backend.github.get_board_id_list("badtoken")


def test_get_rate_limit() -> None:
    """Tests getting the rate limit for an authenticated GitHub request."""
    available, total, reset_time = circfirm.backend.github.get_rate_limit()
    total_rate_limit = 60
    assert available <= total
    assert total == total_rate_limit
    assert reset_time



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""CLI functionality for the current subcommand.

Author(s): Alec Delaney
"""

from typing import Tuple

import click

import circfirm.backend.device
import circfirm.cli


def get_board_info() -> Tuple[str, str]:
    """Get board info via the CLI."""
    circuitpy, _ = circfirm.cli.get_connection_status()
    if not circuitpy:
        raise click.ClickException(
            "Board must be in CIRCUITPY mode in order to detect board information"
        )
    return circfirm.backend.device.get_board_info(circuitpy)


@click.group()
def cli() -> None:
    """Check the information about the currently connected board."""


@cli.command(name="board-id")
def current_board_id() -> None:
    """Get the board ID of the currently connected board."""
    click.echo(get_board_info()[0])


@cli.command(name="version")
def current_version() -> None:
    """Get the CircuitPython version of the currently connected board."""
    click.echo(get_board_info()[1])



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Helpers for tests.

Author(s): Alec Delaney
"""

import os
import pathlib
import platform
import shutil
import time
from typing import Callable, List, TypeVar

import circfirm
import circfirm.backend

_T = TypeVar("_T")


def as_circuitpy(func: Callable[..., _T]) -> Callable[..., _T]:
    """Decorator for running a function with a device connected in CIRCUITPY mode."""  # noqa: D401

    def as_circuitpy_wrapper(*args, **kwargs) -> _T:
        delete_mount_node(circfirm.BOOTOUT_FILE, missing_ok=True)
        delete_mount_node(circfirm.UF2INFO_FILE, missing_ok=True)
        copy_boot_out()
        result = func(*args, **kwargs)
        delete_mount_node(circfirm.BOOTOUT_FILE, missing_ok=True)
        return result

    return as_circuitpy_wrapper


def as_bootloader(func: Callable[..., _T]) -> Callable[..., _T]:
    """Decorator for running a function with a device connected in bootloader mode."""  # noqa: D401

    def as_bootloader_wrapper(*args, **kwargs) -> _T:
        delete_mount_node(circfirm.BOOTOUT_FILE, missing_ok=True)
        delete_mount_node(circfirm.UF2INFO_FILE, missing_ok=True)
        copy_uf2_info()
        result = func(*args, **kwargs)
        delete_mount_node(circfirm.UF2INFO_FILE, missing_ok=True)
        return result

    return as_bootloader_wrapper


def as_not_present(func: Callable[..., _T]) -> Callable[..., _T]:
    """Decorator for running a function without a device connected in either CIRCUITPY or bootloader mode."""  # noqa: D401

    def as_not_present_wrapper(*args, **kwargs) -> _T:
        delete_mount_node(circfirm.BOOTOUT_FILE, missing_ok=True)
        delete_mount_node(circfirm.UF2INFO_FILE, missing_ok=True)
        return func(*args, **kwargs)

    return as_not_present_wrapper


def wait_and_set_bootloader() -> None:
    """Wait then add the boot_out.txt file."""
    time.sleep(2)
    delete_mount_node(circfirm.BOOTOUT_FILE)
    copy_uf2_info()


def set_firmware_version(version: str) -> None:
    """Artificially set the recorded firmware version."""
    bootloader_path = os.path.join(get_mount(), circfirm.BOOTOUT_FILE)

    with open(bootloader_path, encoding="utf-8") as bootfile:
        contents = bootfile.read()

    new_contents = contents.replace("8.0.0-beta.6", version)

    with open(bootloader_path, mode="w", encoding="utf-8") as bootfile:
        bootfile.write(new_contents)


def get_mount() -> str:
    """Get the mounted drive."""
    if platform.system() == "Windows":  # pragma: no cover
        mount_location = "T:\\"
    elif platform.system() == "Darwin":  # pragma: no cover
        mount_location = "/Volumes/TESTMOUNT"
    else:  # pragma: no cover
        mount_location = os.path.join(os.path.curdir, "testmount")
    assert os.path.exists(mount_location)
    assert os.path.isdir(mount_location)
    return (
        mount_location
        if platform.system() == "Windows"
        else os.path.realpath(mount_location)
    )


def get_mount_node(path: str) -> str:
    """Get a file or folder on the mounted drive."""
    mount_location = get_mount()
    return os.path.join(mount_location, path)


def delete_mount_node(path: str, missing_ok: bool = False) -> None:
    """Delete a file on the mounted druve."""
    node_file = get_mount_node(path)
    pathlib.Path(node_file).unlink(missing_ok=missing_ok)


def _copy_text_file(filename: str) -> None:
    """Copy a text file to the mounted test drive."""
    template_bootloader = os.path.join("tests", "assets", filename)
    bootloader_dest = os.path.join(get_mount(), filename)
    shutil.copyfile(template_bootloader, bootloader_dest)


def copy_uf2_info() -> None:
    """Copy a bootloader file to the mounted test drive."""
    _copy_text_file("info_uf2.txt")


def copy_boot_out() -> None:
    """Copy a bootout file to the mounted test drive."""
    _copy_text_file("boot_out.txt")


def copy_firmwares() -> None:
    """Copy firmware files to the app directory."""
    firmware_folder = pathlib.Path("tests/assets/firmwares")
    for board_folder in firmware_folder.glob("*"):
        shutil.copytree(
            board_folder, os.path.join(circfirm.UF2_ARCHIVE, board_folder.name)
        )


def get_board_ids_from_git() -> List[str]:
    """Get a list of board IDs from the sandbox git repository."""
    ports_path = pathlib.Path("tests/sandbox/circuitpython")
    board_paths = ports_path.glob("ports/*/boards/*")
    return sorted(
        [board_path.name for board_path in board_paths if board_path.is_dir()]
    )



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Tests startup functionality.

Author(s): Alec Delaney
"""

import os

import circfirm.backend
import circfirm.startup
import tests.helpers


def test_ensure_dir() -> None:
    """Tests the ensure_dir() function."""
    mount_location = tests.helpers.get_mount()
    test_dir = os.path.join(mount_location, "testdir")

    try:
        circfirm.startup.ensure_dir(test_dir)
        assert os.path.exists(test_dir)
    finally:
        os.rmdir(test_dir)


def test_ensure_file() -> None:
    """Tests the _ensure_file() function."""
    mount_location = tests.helpers.get_mount()
    test_file = os.path.join(mount_location, "testfile")
    try:
        circfirm.startup._ensure_file(test_file)
        assert os.path.exists(test_file)
    finally:
        os.remove(test_file)


def test_ensure_app_setup() -> None:
    """Tests the ensure_app_setup() function."""
    mount_location = tests.helpers.get_mount()
    testfolder = os.path.join(mount_location, "testfolder")
    testfile = os.path.join(testfolder, "testfile")
    circfirm.startup.FOLDER_LIST.append(testfolder)
    circfirm.startup.FILE_LIST.append(testfile)
    try:
        circfirm.startup.ensure_app_setup()
        assert os.path.exists(testfolder)
        assert os.path.exists(testfile)
    finally:
        os.remove(testfile)
        os.rmdir(testfolder)



# SPDX-FileCopyrightText: 2024 Alec Delaney, for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Backend functionality for the working with the cache.

Author(s): Alec Delaney
"""

import os
import pathlib
import re
from typing import Dict, List, Optional, Set, Tuple

import packaging.version
import requests

import circfirm.backend
import circfirm.startup


def get_uf2_filename(board_id: str, version: str, language: str = "en_US") -> str:
    """Get the structured name for a specific board/version CircuitPython."""
    return f"adafruit-circuitpython-{board_id}-{language}-{version}.uf2"


def get_uf2_filepath(
    board_id: str, version: str, language: str = "en_US"
) -> pathlib.Path:
    """Get the path to a downloaded UF2 file."""
    file = get_uf2_filename(board_id, version, language)
    uf2_folder = get_board_folder(board_id)
    return uf2_folder / file


def get_board_folder(board_id: str) -> pathlib.Path:
    """Get the board folder path."""
    return pathlib.Path(circfirm.UF2_ARCHIVE) / board_id


def is_downloaded(board_id: str, version: str, language: str = "en_US") -> bool:
    """Check if a UF2 file is downloaded for a specific board and version."""
    uf2_file = get_uf2_filepath(board_id, version, language)
    return uf2_file.exists()


def download_uf2(board_id: str, version: str, language: str = "en_US") -> None:
    """Download a version of CircuitPython for a specific board."""
    file = get_uf2_filename(board_id, version, language=language)
    uf2_file = get_uf2_filepath(board_id, version, language=language)
    url = f"https://downloads.circuitpython.org/bin/{board_id}/{language}/{file}"
    response = requests.get(url)

    SUCCESS = 200
    if response.status_code != SUCCESS:
        raise ConnectionError(f"Could not download the specified UF2 file:\n{url}")

    uf2_file.parent.mkdir(parents=True, exist_ok=True)
    with open(uf2_file, mode="wb") as uf2file:
        uf2file.write(response.content)


def get_sorted_boards(board_id: Optional[str]) -> Dict[str, Dict[str, Set[str]]]:
    """Get a sorted collection of boards, versions, and languages."""
    boards: Dict[str, Dict[str, Set[str]]] = {}
    for board_folder in sorted(os.listdir(circfirm.UF2_ARCHIVE)):
        versions: Dict[str, List[str]] = {}
        sorted_versions: Dict[str, Set[str]] = {}
        if board_id is not None and board_id != board_folder:
            continue
        board_folder_full = get_board_folder(board_folder)
        for item in os.listdir(board_folder_full):
            version, language = parse_firmware_info(item)
            try:
                version_set = set(versions[version])
                version_set.add(language)
                versions[version] = sorted(version_set)
            except KeyError:
                versions[version] = [language]
        for sorted_version in sorted(
            versions, reverse=True, key=packaging.version.Version
        ):
            sorted_versions[sorted_version] = versions[sorted_version]
        boards[board_folder] = sorted_versions
    return boards


def parse_firmware_info(uf2_filename: str) -> Tuple[str, str]:
    """Get firmware info."""
    regex_match = re.match(circfirm.backend.FIRMWARE_REGEX, uf2_filename)
    if regex_match is None:
        raise ValueError(
            "Firmware information could not be determined from the filename"
        )
    version = regex_match[3]
    language = regex_match[2]
    return version, language
